/* POCSAG decoder */
/* reads demodulated data supplied in HEX by DSP4 at 19200 bps */
/* format: SYNCWORD: DATAWORD DATAWORD ...  CR LF */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <dos.h>
#include "serial.h"
#include <conio.h>
#include <string.h>

// #include <unistd.h>

struct batch
{
    unsigned long sync;
    char rate;
    unsigned long data[16];
    char error[16];
};

extern char *lookup(char *phone);

int main (int argc,char *argv[]);
int nextword(unsigned long *word);
int readbatch(struct batch *bp);
int waitserial(void);

#define	TONE	0
#define	NUM	1
#define ALPHA	2

int func[] = { NUM, NUM, NUM, ALPHA };

char bcd[] = {
	'0',	/* 0000 = 0 */
	'8',	/* 0001 = 8 */
	'4',	/* 0010 = 4 */
	' ',	/* 0011 = C */
	'2',	/* 0100 = 2 */
	'?',	/* 0101 = A */
	'6',	/* 0110 = 6 */
	']',	/* 0111 = E */
	'1',	/* 1000 = 1 */
	'9',	/* 1001 = 9 */
	'5',	/* 1010 = 5 */
	'-',	/* 1011 = D */
	'3',	/* 1100 = 3 */
	'U',	/* 1101 = B */
	'7',	/* 1110 = 7 */
	'['	/* 1111 = F */
	};

#define SYNC	0x7cd215d8
#define ISYNC	0x832dea27
#define	IDLE	0x7a89c197
#define	G	0x0769L			/* x^10+x^9+x^8+x^6+x^5+x^3+1 */

struct batch b;			/* DONDE DE TRABAJA EL BATCH */
int wordptr;

int debug = 0;			/* SIN DEBUG */





#define VERSION 0x0101

#define FALSE           0
#define TRUE            (!FALSE)

#define NOERROR         0       /* No error               */
#define BUFOVFL         1       /* Buffer overflowed      */

#define ESC             0x1B    /* ASCII Escape character */
#define ASCII           0x007F  /* Mask ASCII characters  */
#define SBUFSIZ         0x4000  /* Serial buffer size     */

int            SError          = NOERROR;
int            portbase        = 0;
void           interrupt(*oldvects[2])();

static   char  ccbuf[SBUFSIZ];
unsigned int   startbuf        = 0;
unsigned int   endbuf          = 0;


/* Handle communications interrupts and put them in ccbuf */
void   interrupt com_int(void)
{
    disable();
    if ((inportb(portbase + IIR) & RX_MASK) == RX_ID)
    {
        if (((endbuf + 1) & SBUFSIZ - 1) == startbuf)
            SError = BUFOVFL;

       ccbuf[endbuf++] = inportb(portbase + RXR);
        endbuf &= SBUFSIZ - 1;
    }

    /* Signal end of hardware interrupt */
    outportb(ICR, EOI);
    enable();
}

/* Output a character to the serial port */
int    SerialOut(char x)
{
    long int   timeout = 0x0000FFFFL;

    outportb(portbase + MCR,  MC_INT | DTR | RTS);

    /* Wait for Clear To Send from modem */
    while ((inportb(portbase + MSR) & CTS) == 0)
        if (!(--timeout))
            return (-1);

    timeout = 0x0000FFFFL;

    /* Wait for transmitter to clear */
    while ((inportb(portbase + LSR) & XMTRDY) == 0)
        if (!(--timeout))
            return (-1);

    disable();
    outportb(portbase + TXR, x);
    enable();

    return (0);
}

/* Output a string to the serial port */
void   SerialString(char *string)
{
    while (*string)
       SerialOut(*string++);
}

/* This routine returns the current value in the buffer */
int    getccb(void)
{
    int                res;

    if (endbuf == startbuf)
        return (-1);

    res = (int) ccbuf[startbuf++];
    startbuf %= SBUFSIZ;
    return (res);
}

/* Install our functions to handle communications */
void   setvects(void)
{
    oldvects[0] = getvect(0x0B);
    oldvects[1] = getvect(0x0C);
    setvect(0x0B, com_int);
    setvect(0x0C, com_int);
}

/* Uninstall our vectors before exiting the program */
void   resvects(void)
{
    setvect(0x0B, oldvects[0]);
    setvect(0x0C, oldvects[1]);
}
/* Turn on communications interrupts */
void   i_enable(int pnum)
{
    int                c;

    disable();
    c = inportb(portbase + MCR) | MC_INT;
    outportb(portbase + MCR, c);
    outportb(portbase + IER, RX_INT);
    c = inportb(IMR) & (pnum == COM1 ? IRQ4 : IRQ3);
    outportb(IMR, c);
    enable();
}

/* Turn off communications interrupts */
void   i_disable(void)
{
    int                c;

    disable();
    c = inportb(IMR) | ~IRQ3 | ~IRQ4;
    outportb(IMR, c);
    outportb(portbase + IER, 0);
    c = inportb(portbase + MCR) & ~MC_INT;
    outportb(portbase + MCR, c);
    enable();
}

/* Tell modem that we're ready to go */
void   comm_on(void)
{
    int                c, pnum;

    pnum = (portbase == COM1BASE ? COM1 : COM2);
    i_enable(pnum);
    c = inportb(portbase + MCR) | DTR | RTS;
    outportb(portbase + MCR, c);
}
/* Go off-line */
void   comm_off(void)
{
    i_disable();
    outportb(portbase + MCR, 0);
}

void   initserial(void)
{
    endbuf = startbuf = 0;
    setvects();
    comm_on();
}

void   CloseSerial(void)
{
    comm_off();
    resvects();
}



/* Set the port number to use */
int    SetPort(int Port)
{
    int                Offset, far *RS232_Addr;

    switch (Port)
    { /* Sort out the base address */
      case COM1 : Offset = 0x0000;
		  break;
      case COM2 : Offset = 0x0002;
		  break;
      default   : return (-1);
    }

    RS232_Addr = MK_FP(0x0040, Offset);  /* Find out where the
					    port is.  */
    if (*RS232_Addr == NULL) return (-1);/* If NULL, then port
					    not used. */
    portbase = *RS232_Addr;              /* Otherwise, set
                                            portbase. */

    return (0);
}

/* This routine sets the speed; will accept funny baud rates. */
/* Setting the speed requires that the DLAB be set on.        */
int    SetSpeed(int Speed)
{
    char		c;
    int		divisor;

    if (Speed == 0)            /* Avoid divide by zero */
        return (-1);
    else
	divisor = (int) (115200L/Speed);

    if (portbase == 0)
        return (-1);

    disable();
    c = inportb(portbase + LCR);
    outportb(portbase + LCR, (c | 0x80)); /* Set DLAB   */
    outportb(portbase + DLL, (divisor & 0x00FF));
    outportb(portbase + DLH, ((divisor >> 8) & 0x00FF));
    outportb(portbase + LCR, c);          /* Reset DLAB */
    enable();

    return (0);
}
/* Set other communications parameters */
int    SetOthers(int Parity, int Bits, int StopBit)
{
    int                setting;

    if (portbase == 0)					return (-1);
    if (Bits < 5 || Bits > 8)				return (-1);
    if (StopBit != 1 && StopBit != 2)		return (-1);
    if (Parity != NO_PARITY && Parity != ODD_PARITY && Parity !=
        EVEN_PARITY)                         return (-1);

    setting  = Bits-5;
    setting |= ((StopBit == 1) ? 0x00 : 0x04);
    setting |= Parity;

    disable();
    outportb(portbase + LCR, setting);
    enable();

    return (0);
}

/* Set up the port */
int    SetSerial(int Port, int Speed, int Parity, int Bits, int
                 StopBit)
{
    if (SetPort(Port))                    return (-1);
    if (SetSpeed(Speed))                  return (-1);
    if (SetOthers(Parity, Bits, StopBit)) return (-1);

    return (0);
}

/* Control-Break interrupt handler */
int    c_break(void)
{
    i_disable();
    fprintf(stderr, "\nStill online.\n");

    return(0);
}


int OpenSerial(int channel)
{
    int port=COM1;
    int        speed    = 19200;
    int        parity   = NO_PARITY;
    int        bits     = 8;
    int        stopbits = 1;

    long i;

    if (channel==2) port=COM2; else port=COM1;
    if (SetSerial(port, speed, parity, bits, stopbits) != 0)
    {
	fprintf(stderr, "Serial Port setup error.\n");
	return (99);
    }

    initserial();

    ctrlbrk(c_break);

    return(0);
}


/****

main()
{
    /* The main loop acts as a dumb terminal. We repeatedly check
       the keyboard buffer, and communications buffer. */

    int        c, done  = FALSE;
    OpenSerial();

    fprintf(stdout, "TURBO C TERMINAL\n...You're now in terminal mode, press [ESC] to quit...\n\n");

    do {

	if (kbhit())
	{
	    /* Look for an Escape key */
	    switch (c=getche())
	    {
		case ESC: done = TRUE;  /* Exit program */
			  break;

		/* You may want to handle other keys here... */
	    }
	    SerialOut(c);
	}
	if ((c=getccb()) != -1)
	    fputc(c & ASCII, stdout);

    } while (!done && !SError);

    /* Check for errors */
    switch (SError)
    {
	case NOERROR: fprintf(stderr, "\nbye.\n");
		      CloseSerial();
		      return (0);

	case BUFOVFL: fprintf(stderr, "\nBuffer Overflow.\n");
		      CloseSerial();
		      return (99);

	default:      fprintf(stderr, "\nUnknown Error, SError = %d\n", SError);
		      CloseSerial();
                      return (99);
    }
}





****/












int main (int argc,char *argv[])

{
    int Salida=0;
    unsigned long pager,word;
    int code,i,n,nwords,ascii,bits,type;
    char rate,error;
    char *p;
    char message[256];
    unsigned long mword[128];


    clrscr();
    if (argc < 2) {
	fprintf(stderr,"Usage is: %s <COM#> [<debugflags>]\n",argv[0]);
	return 1;
    }
    printf("DECODIFICADOR POCSAG \r\n");
    printf("BY DSP GROUP OF AMSAT. January 1997 \r\n\r\n");

    if(OpenSerial(atoi(argv[1]))) {
	fprintf(stderr,"%s: cannot open port\n",argv[0]);
	return 2;
    }

/*
    if (!OpenSerial(atoi(argv[1]),19200)) {
	fprintf(stderr,"%s: cannot open port\n",argv[0]);
	return 2;
    }
*/

    if (argc > 2)
	debug = atoi(argv[2]);

    wordptr = 15;
    word = ~0;


    while(!Salida)
    {
	if(kbhit()) { if (getch()==27) Salida=-1; }

	while (word & 0x80000000)		/* message word */
	    if (nextword(&word)) {
		word = ~0L;
		continue;			/* error */
	    }

	if (word == IDLE) {
	    word = ~0L;
	    continue;				/* IDLE word */
	}

	word >>= 11;				/* get rid of CRC */

	pager = ((word & 0xffffc) << 1) + (wordptr >> 1);
	code = word & 3;
	rate = b.rate;
	error = 0;

	/* collect message words for this address */

	for (nwords = 0; nwords < 128; nwords++) {
	    if (nextword(&word) || b.rate != rate) {
		error++;
		break;
	    }

	    if (!(word & 0x80000000))		/* address word */
		break;

	    mword[nwords] = (word >> 11) & 0xfffff;
	}

	/* try to determine type of message */

	if (nwords == 0)
	    type = ALPHA; // type = TONE;

	else {
	    type = func[code];

	    if (code == 0 && nwords > 3 && (mword[nwords - 1] & 0x0f) != 0x03)
		type = ALPHA;
	}

	printf("%c%07ld %d ",(rate == ':'? 'L' : 'H'),pager,code);

	p = message;


	type= ALPHA;  // (Porque lo digo yo)

	switch (type)
	{
	case TONE:
	    printf("[T]");
	    break;

	case NUM:
	    for (n = 0; n < nwords; n++) {
		for (i = 4; i >= 0; i--) {
		    ascii = bcd[(mword[n] >> (4 * i)) & 0x0f];
		    putchar(ascii);
		    if (isdigit(ascii))
			*p++ = ascii;
		}
	    }
	    break;

	case ALPHA:
	    bits = ascii = 0;

	    for (n = 0; n < nwords; n++) {
		  for (i = 19; i >= 0; i--) {
		    ascii >>= 1;
		    if (mword[n] & (1L << i))
			ascii += 64;

		    if (++bits == 7) {

			if (ascii == 0 || ascii == 0x17) {	/* end */
			    if (n != (nwords - 1))
				error++;

			    break;
			}

			if (ascii >= 0x20)
			    putchar(ascii);
			else
			    printf("<%02x>",ascii);

			if (isdigit(ascii))
			    *p++ = ascii;

			bits = ascii = 0;
		    }
		}
	    }
	    break;
	}

	*p = '\0';

	if (error)
	    printf(" [E]");

#ifdef LOOKUP
	if (message[0] == '0' && (p - message) >= 6 && (p - message) <= 10) {
	    if ((p = lookup(message)) != NULL)
		printf(" -> %s",p);
	}
#endif

	printf("\n");
	fflush(stdout);
    }

    CloseSerial();
    return 0;
}

// **************************************************************************
// *
// **************************************************************************

int nextword(unsigned long *word)
{
    if (wordptr == 15) {
	switch (readbatch(&b))
	{
	case 0:					/* ok */
	    wordptr = 0;
	    break;

	case 1:					/* nothing */
	    *word = IDLE;
	    return 1;

	case 2:					/* error */
	    printf("ERROR en el cierre    \n");
	    *word = IDLE;
	    return 1;
	case 3:					/* error */
	    printf("ERROR en el paquete   \n");
	    *word = IDLE;
	    return 1;
	case 4:					/* error */
	    printf("ERROR en syncro       \n");
	    *word = IDLE;
	    return 1;
	case 5:					/* error */
	    printf("ERROR en velocidad    \n");
	    *word = IDLE;
	    return 1;
	}
    } else
	wordptr++;

    /* two subsequent error words in a batch indicates the end of the */
    /* transmission.  let's return an IDLE word for convenience */
    /* (so the levels above assume end of message) and drop this batch */

    if (wordptr < 15 && b.error[wordptr] > 1 && b.error[wordptr + 1] > 1) {
	*word = IDLE;
	wordptr = 15;
	return 0;
    }

    *word = b.data[wordptr];
    return b.error[wordptr];
}

// **************************************************************************
// * LEE UN BATCH
// * ========================================================================
// *
// *	Segn parece el DSP enva un batch completo, que incluye la palabra
// * de sincronismo, y los 8 marcos de 2 veces 32 bit, adems de un carcter
// * con la velocidad entre el sincronismo y los marcos.
// * 	Aqu se transfiere desde el buffer hasta una estructura para
// * la integridad de los datos y en su caso, corregir posibles errores
// * La salida es la estructura corregida con una indicacion de los marcos
// * que pueden presentar un error. en la parte final de la estructura.
// **************************************************************************

int readbatch(struct batch *bp)
{
    int n,c,i;
    char buf[256];
    unsigned long w,g,b;
    unsigned long try,bit;
    static char nothing = 0;

    n = 0;

    while ((c = waitserial()) >= 0 && c != '\n' && n < (sizeof(buf) - 2))
	buf[n++] = c;

    if (c < 0) {
	if (!nothing && (debug & 3))
	    putc('\n',stderr);

	nothing = 1;
	return 1;
    }

/*
    clrscr();
    printf("\r\n--------\r\n");
    for(i=0;i<n;i++) putchar(buf[i]);
    printf("\r\n--------\r\n");
*/

    buf[n++] = c;
    buf[n] = '\0';

    nothing = 0;

    if (debug & 1)
	fprintf(stderr,"%s",buf);

    if (c != '\n')
	return 2;

    n = sscanf(buf,"%lx%c %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx %lx\r\n",
		&bp->sync,&bp->rate,
		&bp->data[0], &bp->data[1], &bp->data[2], &bp->data[3],
		&bp->data[4], &bp->data[5], &bp->data[6], &bp->data[7],
		&bp->data[8], &bp->data[9], &bp->data[10],&bp->data[11],
		&bp->data[12],&bp->data[13],&bp->data[14],&bp->data[15]);

    if (n != 18)
	return 3;

    if (bp->sync != SYNC && bp->sync != ISYNC)
	return 4;

    if (bp->rate != ':' && bp->rate != ';')
	return 5;


    for (n = 0; n < 16; n++)
    {
	/* parity check */

	c = (bp->data[n] >> 16) ^ bp->data[n];
	c = (c >> 8) ^ c;
	c = (c >> 4) ^ c;
	c = (c >> 2) ^ c;
	c = (c >> 1) ^ c;
	bp->error[n] = c & 1;

	if (!bp->error[n]) {
	    /* CRC check */

	    w = bp->data[n] & 0xfffff800;
	    g = G << 21;
	    b = 0x80000000;

	    do
	    {
		if (w & b)
		    w ^= g;

		g >>= 1;
	    } while ((b >>= 1) != 0x400);

	    if ((w ^ bp->data[n]) & 0x7fe) {
		/* CRC error, could be a two-bit error */
		/* attempt to correct it */

		bp->error[n] = '@' - ' ';

		bit = 0xc0000000;

		do
		{
		    try = bp->data[n] ^ bit;

		    w = try & 0xfffff800;
		    g = G << 21;
		    b = 0x80000000;

		    do
		    {
			if (w & b)
			    w ^= g;

			g >>= 1;
		    } while ((b >>= 1) != 0x400);

		    if (((w ^ try) & 0x7fe) == 0) {
			bp->data[n] = try;
			bp->error[n] = '"' - ' ';
			break;
		    }
		} while ((bit >>= 1) != 1);
	    }
	} else {
	    /* parity error, could be single-bit error */
	    /* attempt to correct it */

	    bit = 0x80000000;
	    try = bp->data[n];

	    do
	    {
		w = try & 0xfffff800;
		g = G << 21;
		b = 0x80000000;

		do
		{
		    if (w & b)
			w ^= g;

		    g >>= 1;
		} while ((b >>= 1) != 0x400);

		if (((w ^ try) & 0x7fe) == 0) {
		    bp->data[n] = try;
		    bp->error[n] = '\'' - ' ';
		    break;
		}

		try = bp->data[n] ^ bit;
	    } while (bit >>= 1);
	}

	if (debug & 2)
	    fprintf(stderr,"%08lx%c ",bp->data[n],bp->error[n] + ' ');

	if (debug & 4 || bp->error[n] == '"' - ' ' || bp->error[n] == '\'' - ' ')
	    bp->error[n] = 0;
    }


    if (debug & 2)
	putc('\n',stderr);

    return 0;
}

// **************************************************************************
// * ESPERA UN CARACTER SERIE (YO TENGO ESPERA EN ReadSerial)
// **************************************************************************

int waitserial(void)
{
    int c;
    unsigned long i;
    unsigned long timeout;

    if(kbhit()) {
      if (getch()==27) {
       CloseSerial();
	exit(1);
      }
    }

/*
    for(i=0;i<30000L;i++);
  if ((c=getccb()) != -1)
    fputc(c & ASCII, stdout);
*/

    timeout=10000000L;
    while(((c=getccb()) == -1) && --timeout>0);
    if (!timeout) { return(-1); }
//    fputc(c & ASCII, stdout);
    return(c&0x7f);


/*
    if ((c=getccb()) == -1) for(i=0;i<3000000L;i++);
    if ((c=getccb()) == -1) return(-1);
    fputc(c & ASCII, stdout);
    return(-1);
    return(c&0x7f);
*/


//    if ((c = ReadSerial()) < 0) {
//	usleep(250000);
//		if ((c = ReadSerial()) < 0)
//	    return -1;
//    }
//    return c & 0x7f;

   return -1;
}

// SIEMPRE ALPHA

// bios


